from pydex.dalvik.models.dalvik import (
    DalvikRawItem,
    DalvikHeader,
    DalvikHeaderItem,
    DalvikStringItem,
    DalvikStringData,
    DalvikStringID,
    LazyDalvikString,
    DalvikTypeID,
    DalvikTypeItem,
)



from __future__ import annotations

import asyncio
import struct
import typing
import zlib

from dataclasses import dataclass

from datastream import DeserializingStream, ByteOrder

from pydex.dalvik.models import (
    DalvikHeader,
    DalvikHeaderItem,
    LazyDalvikString,
    DalvikStringID,
    DalvikStringItem,
    DalvikTypeItem,
    DalvikTypeID,
)
from pydex.exc import InvalidDalvikHeader


@dataclass
class DexPool:
    """A container for managing a collection of DexFile instances.

    This class provides a centralized structure for storing and managing multiple DexFile objects,
    allowing for operations that need to interact with or manipulate multiple dex files simultaneously.
    It is particularly useful in scenarios where dex files need to be aggregated for analysis,
    modification, or querying in a unified manner.
    """

    dex_files: list[DexFile]  #: The list of dex files that are managed by this pool.


class DexFile:
    """Represents a Dex file and provides methods to parse and manipulate it.

    This class encapsulates the functionality required to parse and interact with
    the contents of a Dex (Dalvik Executable) file. It provides methods to parse
    the file both synchronously and asynchronously.

    :param data: The raw bytes of the dex file.
    :param no_lazy_load: A flag that indicates whether lazy loading should be disabled.
    """

    def __init__(self, data: bytes, no_lazy_load: bool = False):
        #: The raw bytes of the dex file.
        self.data: bytes = data

        #: A flag that indicates whether lazy loading should be disabled.
        self.no_lazy_load: bool = no_lazy_load

        #: The stream used to read the dex file.
        self.stream: DeserializingStream = DeserializingStream(data, ByteOrder.LITTLE_ENDIAN)

        #: The header of the dex file.
        self.header: DalvikHeaderItem = typing.cast(DalvikHeaderItem, None)

        #: The list of dalvik string items in the dex file.
        self.strings: list[LazyDalvikString | DalvikStringItem] = []

        #: The list of dalvik type items in the dex file.
        self.types: list[DalvikTypeItem] = []

    @classmethod
    def from_path(cls, path: str, no_lazy_load: bool = False) -> DexFile:
        """
        Create a :class:`~pydex.dalvik.DexFile` object from a file path.

        :param path: The path to the dex file.
        :param no_lazy_load: A flag that indicates whether lazy loading should be disabled.
        """

        with open(path, "rb") as f:
            return cls(f.read(), no_lazy_load=no_lazy_load)

    def parse_dex_prologue(self) -> typing.Self:
        """
        Helper function that does misc startup tasks for parsing the dex file.

        :raises InvalidDalvikHeader: If the endian tag is invalid.
        """

        # reset the stream in case it was read from before
        self.stream.seek(0)

        # get the byte order of the dex file
        # it is at 8 + 4 + 20 + 4 + 4 = 40
        endian_tag = int.from_bytes(self.stream.seekpeek(40, 4))

        if endian_tag == 0x12345678:
            self.stream.byteorder = ByteOrder.BIG_ENDIAN
        elif endian_tag == 0x78563412:
            self.stream.byteorder = ByteOrder.LITTLE_ENDIAN
        else:
            raise InvalidDalvikHeader("Invalid endian tag", InvalidDalvikHeader.INVALID_ENDIAN_TAG)

        # parse the header
        self.header = self.parse_header()

        return self

    def parse_dex(self) -> typing.Self:
        """Parse the DEX file.

        This function will attempt to entirely parse this DEX file in one go and fill in all the uninitialized class
        attributes.
        """

        self.parse_dex_prologue()

        # collect all the dalvik string items
        if not self.no_lazy_load:
            self.strings = self.parse_strings()
        else:
            self.strings = self.load_all_strings()

        # collect all the dalvik type items
        self.types = self.parse_types()

        return self

    def parse_header(self) -> DalvikHeaderItem:
        r"""
        Parse the header of the dex file.

        Raises:
            InvalidDalvikHeader: If the parser encounters unexpected values in the header.
        """

        clonestream = self.stream.clone()

        # get the magic bytes
        magic = clonestream.read(8)

        # the magic bytes are not constant, as they contain a version number
        # dex\0x0Axxx\x00, where xxx is the version number(zero padded).
        if magic[:4] != b"dex\x0A":
            raise InvalidDalvikHeader("Invalid magic bytes", InvalidDalvikHeader.INVALID_MAGIC_BYTES)

        if magic[7] != 0:
            raise InvalidDalvikHeader("No terminating NULL character", InvalidDalvikHeader.INVALID_MAGIC_BYTES)

        # get the checksum
        checksum = clonestream.read_uint32()

        # verify the checksum. the checksum applies to the entire file except
        # the magic bytes and this field, so 8 bytes + 4 bytes = 12 bytes to
        # skip.
        if checksum != zlib.adler32(self.data[12:]):
            raise InvalidDalvikHeader(f"Invalid checksum: {checksum}", InvalidDalvikHeader.INVALID_CHECKSUM)

        # get the sha-1 signature bytes(fingerprint)
        signature = clonestream.read(20)

        # get the file size
        file_size = clonestream.read_uint32()

        # get the header size
        header_size = clonestream.read_uint32()

        # according to the spec, this should always be == 0x70
        if header_size != 0x70:
            raise InvalidDalvikHeader(f"Invalid header size: {header_size}", InvalidDalvikHeader.INVALID_HEADER_SIZE)

        # get the endian tag
        endian_tag = clonestream.read_uint32()

        # get the link size
        link_size = clonestream.read_uint32()

        # get the link offset
        link_off = clonestream.read_uint32()

        # get the map offset
        map_off = clonestream.read_uint32()

        # get the string ids size
        string_ids_size = clonestream.read_uint32()

        # get the string ids offset
        string_ids_off = clonestream.read_uint32()

        # get the type ids size
        type_ids_size = clonestream.read_uint32()

        # according to the spec, this should always be <= 0xFFFF
        if type_ids_size >= 0xFFFF:
            raise InvalidDalvikHeader(f"Invalid type ids size: {type_ids_size}", InvalidDalvikHeader.INVALID_TYPES_SIZE)

        # get the type ids offset
        type_ids_off = clonestream.read_uint32()

        # get the proto ids size
        proto_ids_size = clonestream.read_uint32()

        # according to the spec, this should always be <= 0xFFFF
        if proto_ids_size >= 0xFFFF:
            raise InvalidDalvikHeader(
                f"Invalid proto ids size: {proto_ids_size}", InvalidDalvikHeader.INVALID_PROTOS_SIZE
            )

        # get the proto ids offset
        proto_ids_off = clonestream.read_uint32()

        # get the field ids size
        field_ids_size = clonestream.read_uint32()

        # get the field ids offset
        field_ids_off = clonestream.read_uint32()

        # get the method ids size
        method_ids_size = clonestream.read_uint32()

        # get the method ids offset
        method_ids_off = clonestream.read_uint32()

        # get the class defs size
        class_defs_size = clonestream.read_uint32()

        # get the class defs offset
        class_defs_off = clonestream.read_uint32()

        # get the data size
        data_size = clonestream.read_uint32()

        # this field should be divisible by sizeof(uint)
        if data_size % struct.calcsize("I") != 0:
            raise InvalidDalvikHeader(f"Invalid data size: {data_size}", InvalidDalvikHeader.INVALID_DATA_SIZE)

        # get the data offset
        data_off = clonestream.read_uint32()

        return DalvikHeaderItem.from_raw_item(
            DalvikHeader(
                offset=0,
                size=header_size,
                data=self.data[:0x70],
                magic=magic,
                checksum=checksum,
                signature=signature,
                file_size=file_size,
                header_size=header_size,
                endian_tag=endian_tag,
                link_size=link_size,
                link_off=link_off,
                map_off=map_off,
                string_ids_size=string_ids_size,
                string_ids_off=string_ids_off,
                type_ids_size=type_ids_size,
                type_ids_off=type_ids_off,
                proto_ids_size=proto_ids_size,
                proto_ids_off=proto_ids_off,
                field_ids_size=field_ids_size,
                field_ids_off=field_ids_off,
                method_ids_size=method_ids_size,
                method_ids_off=method_ids_off,
                class_defs_size=class_defs_size,
                class_defs_off=class_defs_off,
                data_size=data_size,
                data_off=data_off,
            )
        )

    async def parse_header_async(self) -> DalvikHeaderItem:
        r"""
        Parse the header of the dex file.

        Raises:
            InvalidDalvikHeader: If the parser encounters unexpected values in the header.
        """

        return await asyncio.to_thread(self.parse_header)

    def parse_strings(self) -> list[LazyDalvikString]:
        """Collect all the dalvik string items.

        This function collects all the dalvik string items in this DEX file and returns them as a list of
        :class:`~pydex.dalvik.models.LazyDalvikString`. A clone stream is used so to not alter the DEX
        file stream.
        """

        if self.header is None:
            self.parse_dex_prologue()

        lazy_strings = []
        clonestream = self.stream.clone()

        for i in range(self.header.raw_item.string_ids_size):
            clonestream.seek(self.header.raw_item.string_ids_off + i * 4)
            string_id_off = clonestream.tell()

            string_data_off = clonestream.read_uint32()
            clonestream.seek(string_data_off)

            lazy_strings.append(
                LazyDalvikString(
                    DalvikStringID(
                        offset=string_id_off,
                        size=4,
                        data=self.data[string_id_off : string_id_off + 4],
                        string_data_off=string_data_off,
                        id_number=i,
                    )
                )
            )

        return lazy_strings

    async def parse_strings_async(self) -> list[LazyDalvikString]:
        """Collect all the dalvik string items.

        This function collects all the dalvik string items in this DEX file and returns them as a list of
        :class:`~pydex.dalvik.models.LazyDalvikString`. A clone stream is used so to not alter the
        DEX file stream.
        """

        return await asyncio.to_thread(self.parse_strings)

    def parse_types(self) -> list[DalvikTypeItem]:
        """Collect all the dalvik type items.

        This function collects all the dalvik type items in this DEX file and returns them as a list of
        :class:`~pydex.dalvik.models.DalvikTypeItem`. A clone stream is used so to not alter the DEX
        file stream.
        """

        if self.header is None:
            self.parse_dex_prologue()

        if not self.strings:
            self.strings = self.parse_strings()

        # if there are still no strings, then we can't parse the types.
        # return immediately.
        if not self.strings:
            return []

        types = []
        clonestream = self.stream.clone()

        for i in range(self.header.raw_item.type_ids_size):
            clonestream.seek(self.header.raw_item.type_ids_off + i * 4)
            type_id_off = clonestream.tell()

            descriptor_idx = clonestream.read_uint32()
            descriptor = self.strings[descriptor_idx]

            types.append(
                DalvikTypeItem(
                    DalvikTypeID(
                        offset=type_id_off,
                        size=4,
                        data=self.data[type_id_off : type_id_off + 4],
                        descriptor_idx=descriptor_idx,
                    ),
                    descriptor=descriptor,
                    id_number=i,
                )
            )

        return types

    async def parse_types_async(self) -> list[DalvikTypeItem]:
        """Collect all the dalvik type items asynchronously.

        This function collects all the dalvik type items in this DEX file and returns them as a list of
        :class:`~pydex.dalvik.models.DalvikTypeItem`. A clone stream is used so to not alter the DEX
        file stream.
        """

        return await asyncio.to_thread(self.parse_types)

    def load_all_strings(self) -> list[DalvikStringItem]:
        """Load all the dalvik string items.

        This function invokes the :meth:`~pydex.dalvik.models.LazyDalvikString.load` function for
        all the lazy dalvik strings in the :attr:`strings` attribute. It will also convert every
        model in :attr:`strings` to a loaded :class:`~pydex.dalvik.models.DalvikStringItem` model.
        """

        if not self.strings:
            self.strings = self.parse_strings()

        # if there are still no strings, then we can't load them.
        # return immediately.
        if not self.strings:
            return []

        strings = []

        for lazy_string in self.strings:
            strings.append(lazy_string.load(self.stream))

        return strings

    async def load_all_strings_async(self) -> list[DalvikStringItem]:
        """Load all the dalvik string items asynchronously.

        This function invokes the :meth:`~pydex.dalvik.models.LazyDalvikString.load_async`
        function for all the lazy dalvik strings in the :attr:`strings` attribute. It will
        also convert every model in :attr:`strings` to a loaded
        :class:`~pydex.dalvik.models.DalvikStringItem` model.
        """

        return await asyncio.to_thread(self.load_all_strings)

    def get_string_by_id(self, string_id: int) -> LazyDalvikString:
        """Get a dalvik string by its id.

        Get a dalvik string by its id. The difference between this, and
        ``dex.strings[id]`` is that this method does not require all the strings
        to be collected from the dex file. This method is useful when you only
        need a single string from the dex file and don't want to load the
        entire dex file with :func:`parse_dex`.

        :param string_id: The id of the string to get. IDs are 0-indexed, and are
            assigned in the order they appear in the dex file.
        """

        if self.header is None:
            self.parse_dex_prologue()

        if len(self.strings) > 0:
            return self.strings[string_id]

        clonestream = self.stream.clone()
        clonestream.seek(self.header.raw_item.string_ids_off + string_id * 4)

        string_id_off = clonestream.tell()
        string_data_off = clonestream.read_uint32()

        return LazyDalvikString(
            DalvikStringID(
                offset=string_id_off,
                size=4,
                data=self.data[string_id_off : string_id_off + 4],
                string_data_off=string_data_off,
                id_number=string_id,
            )
        )



from __future__ import annotations

import asyncio
import io
import zipfile
import pathlib
import json

from collections.abc import Generator

from pydex.dalvik.executable import DexPool, DexFile


class Container:
    """
    A class that represents a container file(zip, apk, xapk, apks, etc.) which
    contains either apk files or dex files.
    """

    def __init__(self, path: str):
        #: The path to the container file.
        self.path: pathlib.Path = pathlib.Path(path)


class InMemoryContainer:
    """
    A class that represents an in-memory container which contains either apk
    files or dex files.
    """

    def __init__(self, data: bytes):
        #: The data of the container file.
        self.data: bytes = data


class DexContainer(Container):
    """
    A class that represents a container file(zip, apk, jar) which
    contains dex files.
    """

    def __init__(self, path: str):
        super().__init__(path)

    def enumerate_dex_files(self) -> Generator[str, None, None]:
        """
        Enumerate the dex files in the container file.

        Returns:
            A generator that yields the dex filepaths in the container file.
        """

        raise NotImplementedError()

    def get_dex_data(self, dex_file: str) -> bytes:
        """
        Get the data of the dex file.

        Parameters:
            str dex_file: The filepath of the dex file in the container file.

        Returns:
            The data of the dex file within the container.
        """

        raise NotImplementedError()

    def fetch_dex_files(self) -> DexPool:
        """
        Lazily load the dex files in the container file.

        Returns:
            A DexPool object containing the dex files.
        """

        return DexPool([DexFile(self.get_dex_data(dex_file)) for dex_file in self.enumerate_dex_files()])

    async def fetch_dex_files_async(self) -> DexPool:
        """
        Lazily load the dex files in the container file asynchronously.

        Returns:
            A DexPool object containing the dex files.
        """

        return await asyncio.to_thread(self.fetch_dex_files)


class InMemoryDexContainer(InMemoryContainer):
    """
    A class that represents a container file(zip, apk, jar) which
    contains dex files.
    """

    def __init__(self, data: bytes):
        super().__init__(data)

    def enumerate_dex_files(self) -> Generator[str, None, None]:
        """
        Enumerate the dex files in the container file.

        Returns:
            A generator that yields the dex filepaths in the container file.
        """

        raise NotImplementedError()

    def get_dex_data(self, dex_file: str) -> bytes:
        """
        Get the data of the dex file.

        Parameters:
            str dex_file: The filepath of the dex file in the container file.

        Returns:
            The data of the dex file within the container.
        """

        raise NotImplementedError()

    def fetch_dex_files(self) -> DexPool:
        """
        Lazily load the dex files in the container file.

        Returns:
            A DexPool object containing the dex files.
        """

        return DexPool([DexFile(self.get_dex_data(dex_file)) for dex_file in self.enumerate_dex_files()])

    async def fetch_dex_files_async(self) -> DexPool:
        """
        Lazily load the dex files in the container file asynchronously.

        Returns:
            A DexPool object containing the dex files.
        """

        return await asyncio.to_thread(self.fetch_dex_files)


class InMemoryZipContainer(InMemoryDexContainer):
    """
    A class that represents an in-memory zip file which contains one or more
    dex files.

    Parameters:
        bytes data: The zip file data.
        bool root_only: If True, only the root files in the archive will be
            considered.
    """

    def __init__(self, data: bytes, root_only: bool = False):
        self.root_only = root_only

        super().__init__(data)

    def enumerate_dex_files(self) -> Generator[str, None, None]:
        with zipfile.ZipFile(io.BytesIO(self.data), "r") as zip_file:
            for file in zip_file.namelist():
                file: pathlib.Path = pathlib.Path(file)

                if self.root_only:
                    if file.parent.name != "":
                        continue

                    if file.suffix == ".dex":
                        yield file.name

                    continue

                if file.suffix == ".dex":
                    yield file.name

    def get_dex_data(self, dex_file: str) -> bytes:
        with zipfile.ZipFile(io.BytesIO(self.data), "r") as zip_file:
            with zip_file.open(dex_file, "r") as file:
                return file.read()


class ZipContainer(DexContainer):
    """
    A class that represents a zip file which contains one or more dex files.

    Parameters:
        str path: The path to the zip file.
        bool root_only: If True, only the root files in the archive will be
            considered.
    """

    def __init__(self, path: str, root_only: bool = False):
        self.root_only = root_only
        self.internal_container = InMemoryZipContainer(pathlib.Path(path).read_bytes())

        super().__init__(path)

    def enumerate_dex_files(self) -> Generator[str, None, None]:
        return self.internal_container.enumerate_dex_files()

    def get_dex_data(self, dex_file: str) -> bytes:
        return self.internal_container.get_dex_data(dex_file)


class InMemoryMultiAPKContainer(InMemoryContainer):
    """
    A class that represents an in-memory apk container which contains one or
    more apk files.

    Parameters:
        bytes data: The apk container file data.
    """

    def __init__(self, data: bytes):
        super().__init__(data)

    def get_base_apk(self) -> bytes:
        """
        Get the base apk file as a bytes object.

        Returns:
            The file data of the base apk file in the container.
        """

        raise NotImplementedError()

    def fetch_dex_files(self, root_only: bool = False) -> DexPool:
        """
        Lazily load the dex files in the base apk file.

        Parameters:
            bool root_only: If True, only the root files in the archive will
                be considered.

        Returns:
            A DexPool object containing the dex files.
        """

        zip_container = InMemoryZipContainer(self.get_base_apk(), root_only)

        return zip_container.fetch_dex_files()

    async def fetch_dex_files_async(self, root_only: bool = False) -> DexPool:
        """
        Lazily load the dex files in the base apk file asynchronously.

        Parameters:
            bool root_only: If True, only the root files in the archive will
                be considered.

        Returns:
            A DexPool object containing the dex files.
        """

        return await asyncio.to_thread(self.fetch_dex_files, root_only)


class MultiAPKContainer(Container):
    """
    A class that represents a container file which contains one or more apk
    files.

    Parameters:
        str path: The path to the apk container file.
    """

    def __init__(self, path: str):
        super().__init__(path)

    def get_base_apk(self) -> bytes:
        """
        Get the base apk file as a bytes object.

        Returns:
            The file data of the base apk file in the container
        """

        raise NotImplementedError()

    def fetch_dex_files(self, root_only: bool = False) -> DexPool:
        """
        Lazily load the dex files in the base apk file.

        Parameters:
            bool root_only: If True, only the root files in the archive will
                be considered.

        Returns:
            A DexPool object containing the dex files.
        """

        zip_container = InMemoryZipContainer(self.get_base_apk(), root_only)

        return zip_container.fetch_dex_files()

    async def fetch_dex_files_async(self, root_only: bool = False) -> DexPool:
        """
        Lazily load the dex files in the base apk file asynchronously.

        Parameters:
            bool root_only: If True, only the root files in the archive will
                be considered.

        Returns:
            A DexPool object containing the dex files.
        """

        return await asyncio.to_thread(self.fetch_dex_files, root_only)


class InMemoryXAPKContainer(InMemoryMultiAPKContainer):
    """
    A class that represents an in-memory xapk container file which contains one
    or more apk files. These conatiner files contain various apk files, whose
    purpose is noted in the ``manifest.json`` file. This class parses the
    ``manifest.json`` file to determine the base apk file, which is annotated by
    the ``base`` field in the ``split_apks`` object.

    Parameters:
        bytes data: The xapk container file data.
    """

    def __init__(self, data: bytes):
        super().__init__(data)

    def get_base_apk(self) -> bytes:
        with zipfile.ZipFile(io.BytesIO(self.data), "r") as zip_file:
            for file in zip_file.namelist():
                file: pathlib.Path = pathlib.Path(file)

                if file.name != "manifest.json":
                    continue

                with zip_file.open(str(file), "r") as manifest_file:
                    manifest = json.load(manifest_file)

                    for split_apk in manifest["split_apks"]:
                        if split_apk["id"] != "base":
                            continue

                        with zip_file.open(split_apk["file"], "r") as base_apk_file:
                            return base_apk_file.read()
                    else:
                        raise ValueError("Base apk not found in manifest file.")
            else:
                raise FileNotFoundError("Manifest file not found in xapk file.")


class XAPKContainer(MultiAPKContainer):
    """
    A class that represents a xapk file which contains one or more apk files.
    These conatiner files contain various apk files, whose purpose is noted in
    the ``manifest.json`` file. This class parses the ``manifest.json`` file to
    determine the base apk file, which is annotated by the ``base`` field in the
    ``split_apks`` object.

    Parameters:
        str path: The path to the xapk file.
    """

    def __init__(self, path: str):
        super().__init__(path)

    def get_base_apk(self) -> bytes:
        return InMemoryXAPKContainer(pathlib.Path(self.path).read_bytes()).get_base_apk()


class InMemoryAPKSContainer(InMemoryMultiAPKContainer):
    """
    A class that represents an in-memory apks container file which contains one
    or more apk files. This is also known as a split-apk. The base apk in these
    containers is usually named ``base.apk``.

    Parameters:
        bytes data: The apks container file data.
    """

    def __init__(self, data: bytes):
        super().__init__(data)

    def get_base_apk(self) -> bytes:
        with zipfile.ZipFile(io.BytesIO(self.data), "r") as zip_file:
            for file in zip_file.namelist():
                file: pathlib.Path = pathlib.Path(file)

                if file.name != "base.apk":
                    continue

                with zip_file.open(str(file), "r") as base_apk_file:
                    return base_apk_file.read()
            else:
                raise FileNotFoundError("Base apk not found in apks file.")


class APKSContainer(MultiAPKContainer):
    """
    A class that represents an apks file which contains one or more apk files.
    This is also known as a split-apk. The base apk in these containers is
    usually named ``base.apk``.

    Parameters:
        str path: The path to the apks file.
    """

    def __init__(self, path: str):
        super().__init__(path)

    def get_base_apk(self) -> bytes:
        return InMemoryAPKSContainer(pathlib.Path(self.path).read_bytes()).get_base_apk()


class JarContainer(ZipContainer):
    """
    A class that represents a jar file which contains one or more dex files.

    Parameters:
        str path: The path to the jar file.
    """

    def __init__(self, path: str):
        super().__init__(path)


class APKContainer(ZipContainer):
    """
    A class that represents an apk file which contains one or more dex files.
    Only the root files in the archive will be considered.

    Parameters:
        str path: The path to the apk file.
    """

    def __init__(self, path: str):
        super().__init__(path, root_only=True)



from __future__ import annotations

import asyncio

from dataclasses import dataclass

from datastream import ByteOrder, DeserializingStream

from pydex.util import sizeof_uleb128
from pydex.exc import InvalidDalvikHeader


@dataclass
class DalvikRawItem:
    """
    A dataclass that represents a low-level item in a dex file.
    """

    #: The offset of the item in the dex file.
    offset: int

    #: The size of the item in the dex file.
    size: int

    #: The raw data of the item.
    data: bytes


@dataclass
class DalvikHeader(DalvikRawItem):
    """
    A dataclass that represents the raw header of a dex file.

    .. admonition:: Source
        :class: seealso

        `dex_format::header_item <https://source.android.com/docs/core/runtime/dex-format#header-item>`_
    """

    #: Magic value.
    magic: bytes  # 8 bytes

    #: Adler32 checksum of the rest of the file (everything but magic and this field).
    #: Used to detect file corruption.
    checksum: int  # 4 bytes

    #: SHA-1 hash of the rest of the file (everything but magic, checksum, and this field).
    #: Used to uniquely identify the file.
    signature: bytes  # 20 bytes

    #: Size of the entire file (including the header), in bytes.
    file_size: int  # 4 bytes

    #: Header size in bytes. This value should be ``0x70``.
    header_size: int  # 4 bytes

    #: Endian tag. The value should be either ``0x12345678`` or ``0x78563412``.
    endian_tag: int  # 4 bytes

    #: Size of the link section.
    link_size: int  # 4 bytes

    #: Offset from the start of the file to the link section.
    link_off: int  # 4 bytes

    #: Offset from the start of the file to the map section.
    map_off: int  # 4 bytes

    #: Size of the string identifiers list.
    string_ids_size: int  # 4 bytes

    #: Offset from the start of the file to the string identifiers list.
    string_ids_off: int  # 4 bytes

    #: Size of the type identifiers list.
    type_ids_size: int  # 4 bytes

    #: Offset from the start of the file to the type identifiers list.
    type_ids_off: int  # 4 bytes

    #: Size of the prototype identifiers list.
    proto_ids_size: int  # 4 bytes

    #: Offset from the start of the file to the prototype identifiers list.
    proto_ids_off: int  # 4 bytes

    #: Size of the field identifiers list.
    field_ids_size: int  # 4 bytes

    #: Offset from the start of the file to the field identifiers list.
    field_ids_off: int  # 4 bytes

    #: Size of the method identifiers list.
    method_ids_size: int  # 4 bytes

    #: Offset from the start of the file to the method identifiers list.
    method_ids_off: int  # 4 bytes

    #: Size of the class definitions list.
    class_defs_size: int  # 4 bytes

    #: Offset from the start of the file to the class definitions list.
    class_defs_off: int  # 4 bytes

    #: Size of the data section.
    data_size: int  # 4 bytes

    #: Offset from the start of the file to the data section.
    data_off: int  # 4 bytes


@dataclass
class DalvikHeaderItem:
    """
    A dataclass that represents the high-level header of a dex file.
    """

    #: The raw header item.
    raw_item: DalvikHeader

    #: The version of the dex file.
    version: int

    #: The adler32 checksum of the dex file.
    checksum: int

    #: The unique SHA-1 fingerprint of the dex file.
    signature: bytes

    #: The size of the dex file in bytes.
    file_size: int

    #: The byte order of the dex file.
    byte_order: ByteOrder

    @classmethod
    def from_raw_item(cls, raw_item: DalvikHeader) -> DalvikHeaderItem:
        """
        Create a DalvikHeaderItem from a DalvikRawItem.

        Args:
            DalvikHeader raw_item: The DalvikRawItem that contains the primitive data for this item.

        Raises:
            InvalidDalvikHeader: If ``endian_tag`` is invalid.
        """

        version = int(raw_item.magic[4:7])
        checksum = raw_item.checksum
        signature = raw_item.signature
        file_size = raw_item.file_size

        if raw_item.endian_tag == 0x12345678:
            byte_order = ByteOrder.LITTLE_ENDIAN
        elif raw_item.endian_tag == 0x78563412:
            byte_order = ByteOrder.BIG_ENDIAN
        else:
            raise InvalidDalvikHeader("Invalid endian tag", InvalidDalvikHeader.INVALID_ENDIAN_TAG)

        return cls(raw_item, version, checksum, signature, file_size, byte_order)


@dataclass
class DalvikStringID(DalvikRawItem):
    """
    A dataclass that represents the raw string id of a dex file. Aligns to 4 bytes.

    .. admonition:: Source
        :class: seealso

        `dex_format::string_id_item <https://source.android.com/docs/core/runtime/dex-format#string-item>`_
    """

    #: Offset from the start of the file to the string data.
    string_data_off: int  # 4 bytes

    #: The index of the string in the string table. This field is not part of
    #: the dex file format.
    id_number: int


@dataclass
class DalvikStringData(DalvikRawItem):
    """
    A dataclass that represents the raw string data of a dex file.

    .. admonition:: Source
        :class: seealso

        `dex_format::string_data_item <https://source.android.com/docs/core/runtime/dex-format#string-data-item>`_
    """

    #: Size of the string in UTF-16 code units.
    utf16_size: int  # uleb128

    #: The raw string data.
    string_data: bytes  # mutf8 encoded


@dataclass
class DalvikStringItem:
    """
    A dataclass that represents a high-level string data item in a dex file.
    """

    #: The raw string data item.
    raw_item: DalvikStringData

    #: The raw string id item.
    string_id: DalvikStringID

    _value: str = ""

    @classmethod
    def from_raw_item(cls, raw_item: DalvikStringData, string_id: DalvikStringID) -> DalvikStringItem:
        """
        Create a DalvikStringItem from a DalvikStringID and DalvikStringData.

        Args:
            DalvikStringData raw_item: The DalvikStringData that will contain the data of this item.
            DalvikStringID string_id: The DalvikStringID.
        """

        return cls(raw_item, string_id)

    @property
    def value(self) -> str:
        """
        Get the value of the string decoded as MUTF-8.

        Raises:
            ValueError: If an invalid MUTF-8 sequence is encountered.

        Returns:
            The MUTF-8 decoded string.
        """
        # check if the value is already decoded
        if self._value:
            return self._value

        data = bytearray(self.raw_item.string_data)
        decoded = ""

        while len(data) > 0:
            b = data.pop(0)

            if b & 0x80 == 0x00:
                decoded += chr(b)
            elif b & 0xE0 == 0xC0:
                decoded += chr(((b & 0x1F) << 6) | (data.pop(0) & 0x3F))
            elif b & 0xF0 == 0xE0:
                decoded += chr(((b & 0x0F) << 12) | ((data.pop(0) & 0x3F) << 6) | (data.pop(0) & 0x3F))
            else:
                raise ValueError("Invalid MUTF-8 sequence")

        self._value = decoded

        return decoded

    @value.setter
    def value(self, value: str):
        """
        Set the value of the dalvik string encoded as MUTF-8.

        Raises:
            ValueError: If an invalid codepoint is encountered.

        Args:
            str value: The string value to set.
        """
        encoded = bytearray()

        for char in value:
            codepoint = ord(char)

            if codepoint <= 0x7F:
                encoded.append(codepoint)
            elif codepoint <= 0x7FF:
                encoded.append(0xC0 | ((codepoint >> 6) & 0x1F))
                encoded.append(0x80 | (codepoint & 0x3F))
            elif codepoint <= 0xFFFF:
                encoded.append(0xE0 | ((codepoint >> 12) & 0x0F))
                encoded.append(0x80 | ((codepoint >> 6) & 0x3F))
                encoded.append(0x80 | (codepoint & 0x3F))
            else:
                raise ValueError("Invalid codepoint")

        self.raw_item.string_data = bytes(encoded)
        self.raw_item.utf16_size = len(encoded)

    async def get_value_async(self) -> str:
        """
        Get the value of the string decoded as MUTF-8 asynchronously.

        Returns:
            The MUTF-8 decoded string.
        """

        return await asyncio.to_thread(self.__class__.value.fget)

    async def set_value_async(self, value: str):
        """
        Set the value of the dalvik string encoded as MUTF-8 asynchronously.

        Args:
            str value: The string value to set.
        """

        return await asyncio.to_thread(self.__class__.value.fset, value)

    def __str__(self) -> str:
        return self.value


@dataclass
class LazyDalvikString:
    """
    A dataclass that represents a dalvik string which can be loaded at any time.
    """

    string_id: DalvikStringID

    def load(self, stream: DeserializingStream) -> DalvikStringItem:
        """
        Load the string from the stream.

        Args:
            DeserializingStream stream: The DeserializingStream to read from.
        Returns:
            A loaded DalvikStringItem.
        """

        # save the current position so to not disrupt the processing flow.
        # we should NOT use stream.clone() here as this will copy the entire
        # stream for every string item which is not efficient.
        pos = stream.tell()
        stream.seek(self.string_id.string_data_off)

        offset = stream.tell()
        utf16_size = stream.read_uleb128()
        data = stream.read(utf16_size)

        leb128_size = sizeof_uleb128(utf16_size)
        item_data = stream.seekpeek(self.string_id.string_data_off, leb128_size + utf16_size)

        # restore the position
        stream.seek(pos)

        return DalvikStringItem(
            DalvikStringData(offset, sizeof_uleb128(utf16_size) + len(data), item_data, utf16_size, data),
            self.string_id,
        )

    async def load_async(self, stream: DeserializingStream) -> DalvikStringItem:
        """
        Load the string from the stream asynchronously.
        Args:
            DeserializingStream stream: The DeserializingStream to read from.

        Returns: A loaded DalvikStringItem.
        """

        return await asyncio.to_thread(self.load, stream)


@dataclass
class DalvikTypeID(DalvikRawItem):
    """
    A dataclass that represents the raw type id of a dex file.

    .. admonition:: Source
        :class: seealso

        `dex_format::type_id_item <https://source.android.com/docs/core/runtime/dex-format#type-id-item>`_
    """

    #: Index into the ``string_ids`` list for the descriptor string of this type.
    descriptor_idx: int  # 4 bytes


@dataclass
class DalvikTypeItem:
    """
    A dataclass that represents a high-level type item in a dex file.
    """

    #: The raw type id item.
    raw_item: DalvikTypeID

    #: The raw string item.
    descriptor: DalvikStringItem | LazyDalvikString

    #: The index number of this type
    id_number: int

    @classmethod
    def from_raw_item(
        cls, raw_item: DalvikTypeID, strings: list[DalvikStringItem | LazyDalvikString], id_number: int
    ) -> DalvikTypeItem:
        """
        Create a DalvikTypeItem from a DalvikTypeID

        Args:
            DalvikTypeID raw_item: The DalvikTypeID that will contain the data of this item.
            list[DalvikStringItem | LazyDalvikString] strings: The list of string items.
            int id_number: The index number of this type
        """

        return cls(raw_item, strings[raw_item.descriptor_idx], id_number)

    def __str__(self) -> str:
        return self.descriptor.value



def sizeof_uleb128(value: int) -> int:
    """
    Get the size of an unsigned LEB128 value.
    """
    size = 0

    while True:
        size += 1
        value >>= 7

        if value == 0:
            break

    return size


def sizeof_sleb128(value: int) -> int:
    """
    Get the size of a signed LEB128 value.
    """
    size = 0

    if value >= 0:
        while value > 0x3F:
            size += 1
            value %= 0x100000000
            value >>= 7

        size += 1
    else:
        while value < -0x40:
            size += 1
            value >>= 7

        size += 1

    return size



class InvalidDalvikHeader(Exception):
    """Exception raised when the parser encounters an invalid Dalvik header.

    A number of codes exist as class attributes in this class which represent
    the reason for the exception. These codes exist so that multiple exception
    classes don't have to be created for each possible reason.
    """

    #: The magic bytes of the header are invalid.
    INVALID_MAGIC_BYTES = 0

    #: The checksum of the header does not match the calculated checksum.
    INVALID_CHECKSUM = 1

    #: The endian tag of the header representing byte order is invalid.
    INVALID_ENDIAN_TAG = 2

    #: The size of the header is not ``0x70``.
    INVALID_HEADER_SIZE = 3

    #: The ``proto_ids_size`` is greater than or equal to ``0xFFFF``.
    INVALID_PROTOS_SIZE = 4

    #: The ``type_ids_size`` is greater than or equal to ``0xFFFF``.
    INVALID_TYPES_SIZE = 5

    #: The ``data_size`` is not divisible by the size of a word.
    INVALID_DATA_SIZE = 6

    def __init__(self, message: str, code: int):
        #: The message of the exception.
        self.message = message

        #: The code of the exception.
        self.code = code

        super().__init__(self.message)
